#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>

#define TAPE_0_PULSE 416
#define TAPE_1_PULSE 208

#define TAPE_DECODE_0_MIN (TAPE_1_PULSE+(TAPE_1_PULSE/5))
#define TAPE_DECODE_0_MAX (TAPE_0_PULSE*2)
#define TAPE_DECODE_1_MIN (TAPE_1_PULSE/4)
#define TAPE_DECODE_1_MAX (TAPE_DECODE_0_MIN)

#define TIME_TO_BIT(t) ((t<TAPE_DECODE_1_MIN)?-1:((t<TAPE_DECODE_0_MIN)?1:0))


// This is used by the "tapsections" function. It returns the next time
// value from an ort buffer, or -1 for invalid (or if an existing tapsection
// is found)
static int next_ort_value(unsigned char *ortdata, int *offset, int end, int *anytapsec)
{
    int val = 0, skip;

    while ((*offset) < end)
    {
        switch (ortdata[*offset])
        {
        case 0xfc:
            if ((*offset) >= (end - 1)) return 0;
            val = ortdata[(*offset) + 1];
            (*offset) += 2;
            return val;

        case 0xfd:
            if ((*offset) >= (end - 2)) return 0;
            val = (ortdata[(*offset) + 1] << 8) | ortdata[(*offset) + 2];
            (*offset) += 3;
            return val;

        case 0xfe:
            return 0;

        case 0xff:
            if ((*offset) >= (end - 2)) return 0;
            skip = (ortdata[(*offset) + 1] << 8) | ortdata[(*offset) + 2];
            (*offset) += 3;
            if ((*offset) >= (end - skip)) return 0;
            (*offset) += skip;
            (*anytapsec) = 1;
            break;

        default:
            return ortdata[(*offset)++];
        }
    }

    return 0;
}

enum
{
    TAPSEC_STATE_SEARCH = 0,
    TAPSEC_STATE_LEADER,
    TAPSEC_STATE_FINDHEADER,
    TAPSEC_STATE_HEADER,
    TAPSEC_STATE_FILENAME,
    TAPSEC_STATE_FINDDATA,
    TAPSEC_STATE_DATA
};

static int validbyte(int accum)
{
    int i, parity = 0;
    if ((accum & 3) != 1) return 0;
    for (i = 4; i != 0x800; i <<= 1)
        if (accum&i) parity ^= 1;
    return parity;
}

// This converts oric standard encoded waveforms to decoded TAP sections
static int tapsections(unsigned char *ortbuf, int ortbuflen, unsigned char *scratch, int slow)
{
    int ort_offs, tapebit, ort_val, cyccount;
    int accum, state, thisbit, leaderstart = 0, leadercount = 0;
    int last_offsets[20], i, bitcount, tapbytes = 0;
    int data_bytes = 0, slow0s = 0, slow1s = 0;
    int anytapsec = 0;

    ort_offs = 4;
    tapebit = ortbuf[ort_offs++];
    cyccount = 0;
    state = TAPSEC_STATE_SEARCH;
    accum = 0;
    bitcount = 0;

    for (i = 0; i<20; i++)
        last_offsets[i] = 5;

    // Loop through the ORT data
    while (ort_offs < ortbuflen)
    {
        // Shuffle the offset array
        for (i = 0; i<19; i++)
            last_offsets[i] = last_offsets[i + 1];
        last_offsets[19] = ort_offs;

        ort_val = next_ort_value(ortbuf, &ort_offs, ortbuflen, &anytapsec);
        // Invalid ORT data?
        if (ort_val < 1) return ortbuflen;

        // Count these cycles
        cyccount += ort_val;

        // Invert the tape input
        tapebit ^= 1;

        // Rising edge?
        if (tapebit)
        {
            // smaller than 100? filter it out
            if (cyccount < 100)
                continue;

            // Convert to 1/0
            thisbit = TIME_TO_BIT(cyccount);

            // Seen a value that is outside the bounds of Oric standard encoding?
            if (thisbit == -1)
            {
                //        if ((state != TAPSEC_STATE_SEARCH) || (slow0s>0) || (slow1s>0))
                //        {
                //          if (cyccount < TAPE_DECODE_1_MIN) printf("%d is too fast\n", cyccount);
                //          if (cyccount > TAPE_DECODE_0_MAX) printf("%d is too slow\n", cyccount);
                //          fflush(stdout);
                //        }
                state = TAPSEC_STATE_SEARCH;
                cyccount = 0;
                accum = 0;
                slow1s = 0;
                slow0s = 0;
            }
            else
            {
                // Slow mode?
                if (slow)
                {
                    if (thisbit == 1)
                    {
                        slow1s++;
                        slow0s = 0;

                        if (slow1s != 2)
                        {
                            if (slow1s >= 8)
                                slow1s = 0;
                            cyccount = 0;
                            continue;
                        }
                    }
                    else
                    {
                        slow0s++;
                        slow1s = 0;

                        if (slow0s != 2)
                        {
                            if (slow0s >= 4)
                                slow0s = 0;
                            cyccount = 0;
                            continue;
                        }
                    }
                }

                // Accumulate this bit
                accum = (accum >> 1) | (thisbit << 10);

                // What are we doing?
                switch (state)
                {
                case TAPSEC_STATE_SEARCH:
                    // Look for 0x058 (0x16 encoded with start bit and parity)
                    if ((accum & 0x7fe) == 0x58)
                    {
                        //              printf("Seen 0x16 @ %d\n", ort_offs);
                        //              fflush(stdout);
                        leaderstart = last_offsets[8];
                        leadercount = 1;
                        state = TAPSEC_STATE_LEADER;
                        bitcount = 0;
                    }
                    else if (validbyte(accum))
                    {
                        //              printf("Saw %02X @ %d\n", (accum>>2)&0xff, ort_offs);
                        //              fflush(stdout);
                    }
                    break;

                case TAPSEC_STATE_LEADER:
                    bitcount = (bitcount + 1) % 14;
                    if ((bitcount == 13) && validbyte(accum)) bitcount = 0;
                    if (!bitcount)
                    {
                        //              printf("Byte @ %d is %03x\n", ort_offs, accum&0x7fe);
                        //              fflush(stdout);
                        if ((accum & 0x7fe) != 0x58)
                        {
                            //                printf("Which is NOT 0x58 :-(\n");
                            state = TAPSEC_STATE_SEARCH;
                            cyccount = 0;
                            break;
                        }

                        leadercount++;
                        if (leadercount == 4)
                        {
                            state = TAPSEC_STATE_FINDHEADER;
                            break;
                        }
                    }
                    break;

                case TAPSEC_STATE_FINDHEADER:
                    bitcount = (bitcount + 1) % 14;
                    if ((bitcount == 13) && validbyte(accum)) bitcount = 0;
                    if (!bitcount)
                    {
                        //              printf("Byte @ %d is %03x\n", ort_offs, accum&0x7fe);
                        //              fflush(stdout);
                        if ((accum & 0x7fe) == 0x490) // 0x24 fully encoded
                        {
                            //                printf("At the header...\n");
                            //                fflush(stdout);
                            scratch[0] = 0x16;
                            scratch[1] = 0x16;
                            scratch[2] = 0x16;
                            scratch[3] = 0x24;
                            tapbytes = 4;
                            state = TAPSEC_STATE_HEADER;
                        }
                    }
                    break;

                case TAPSEC_STATE_HEADER:
                    bitcount = (bitcount + 1) % 14;
                    if ((bitcount == 13) && validbyte(accum)) bitcount = 0;
                    if (!bitcount)
                    {
                        //              printf("Byte @ %d is %03x (%02X)\n", ort_offs, accum&0x7fe, (accum>>2)&0xff);
                        //              fflush(stdout);
                        scratch[tapbytes++] = (accum >> 2) & 0xff;
                        if (tapbytes == 13)
                        {
                            int load_start, load_end;

                            load_start = (scratch[10] << 8) | scratch[11];
                            load_end = (scratch[8] << 8) | scratch[9];

                            //                printf("Load start = %04X, Load end = %04X\n", load_start, load_end);
                            //                fflush(stdout);

                            if (load_end <= load_start)
                            {
                                state = TAPSEC_STATE_SEARCH;
                                cyccount = 0;
                                break;
                            }

                            data_bytes = (load_end - load_start) + 1;
                            state = TAPSEC_STATE_FILENAME;

                            //                printf("Lets try and load %d bytes!\n", data_bytes);
                            //                fflush(stdout);
                        }
                    }
                    break;

                case TAPSEC_STATE_FILENAME:
                    bitcount = (bitcount + 1) % 14;
                    if ((bitcount == 13) && validbyte(accum)) bitcount = 0;
                    if (!bitcount)
                    {
                        //              printf("Filename byte = %03x (%02x) %c\n", accum&0x7fe, (accum>>2)&0xff, (accum>>2)&0xff);
                        //              fflush(stdout);
                        scratch[tapbytes++] = (accum >> 2) & 0xff;
                        if (scratch[tapbytes - 1] == 0)
                        {
                            state = TAPSEC_STATE_FINDDATA;
                        }
                    }
                    break;

                case TAPSEC_STATE_FINDDATA:
                    // Look for a valid byte (start bits, valid parity)
                    if (validbyte(accum))
                    {
                        //              printf("Data starts @ %d\n", ort_offs);
                        //              fflush(stdout);
                        scratch[tapbytes++] = (accum >> 2) & 0xff;
                        data_bytes--;
                        bitcount = 0;
                        state = TAPSEC_STATE_DATA;
                    }
                    break;

                case TAPSEC_STATE_DATA:
                    bitcount = (bitcount + 1) % 14;
                    if ((bitcount == 13) && validbyte(accum)) bitcount = 0;
                    if (!bitcount)
                    {
                        scratch[tapbytes++] = (accum >> 2) & 0xff;
                        data_bytes--;
                        //              printf("%d to go..\n", data_bytes);
                        //              fflush(stdout);

                        // ??
                        if (data_bytes < 0)
                        {
                            state = TAPSEC_STATE_SEARCH;
                            break;
                        }

                        if (data_bytes > 0)
                            break;

                        if ((!anytapsec) && (leaderstart < 512))
                            leaderstart = 5;

                        // Got a whole standard oric tap section!
                        ortbuf[leaderstart++] = 0xff;
                        ortbuf[leaderstart++] = (tapbytes >> 8) & 0xff;
                        ortbuf[leaderstart++] = (tapbytes & 0xff);
                        memcpy(&ortbuf[leaderstart], scratch, tapbytes);
                        leaderstart += tapbytes;
                        memmove(&ortbuf[leaderstart], &ortbuf[ort_offs], ortbuflen - ort_offs);
                        ortbuflen -= (ort_offs - leaderstart);
                        ort_offs = leaderstart;
                        state = TAPSEC_STATE_SEARCH;
                        accum = 0;
                        anytapsec = 1;

                        //              printf("Back to scanning...\n");
                        //              fflush(stdout);
                    }
                    break;
                }
            }

            cyccount = 0;
        }
    }

    return ortbuflen;
}

int main() {
    unsigned char scratch[65536];
    unsigned char * buffer;
    int len;

    FILE *f;

    char fname[] = "C:\\Users\\engui\\OneDrive\\Oric\\Torreoscura\\torre2.ort";
    char fout[] = "C:\\Users\\engui\\OneDrive\\Oric\\Torreoscura\\torre2.tap";

    // First make sure the image file exists
    f = fopen(fname, "rb");
    if (!f) {
        printf("Cannot find input file");
        return -1;
    }

    // Get the image size
    fseek(f, 0, SEEK_END);
    len = (int)ftell(f);
    fseek(f, 0, SEEK_SET);

    // Allocate memory for the tape image and read it in
    buffer = (unsigned char *)malloc(sizeof(char)*(len + 1));
    if (!buffer)
    {
        fclose(f);
        printf("Memory alloc error");
        return -1;
    }

    if (fread(buffer, len, 1, f) != 1)
    {
        fclose(f);
        printf("Read error");
        return -1;
    }

    fclose(f);

    int retval = tapsections(buffer, len, scratch, 0);
    printf("Done. Return value is %d\n", retval);

    f = fopen(fout, "wb");
    fwrite(scratch, retval, 1, f);
    fclose(f);

    //system("PAUSE");

}
